import html
import re

import colorama
from termcolor import colored

from ekphrasis.classes.exmanager import ExManager


class Tokenizer:
    social_pipeline = [
        "EMOJI_UCS4", "URL", "TAG", "EMAIL", "USER", "HASHTAG",
        "CASHTAG", "PHONE", "PERCENT", "MONEY", "DATE", "TIME",
        "ACRONYM", "LTR_FACE", "RTL_FACE", "CENSORED", "EMPHASIS",
        "REST_EMOTICONS", "NUMBER", "WORD", "EASTERN_EMOTICONS",
    ]
    default_pipeline = social_pipeline

    def __init__(self, pipeline=None, lowercase=False, verbose=False,
                 debug=False):
        """
        Args:
            pipeline (list): list of terms to use for tokenization.
                Each term, is a key from the dict of regexes `expressions.txt`.
                Order matters!
            lowercase (bool): set to True in order to lowercase the text
            verbose (bool): set to True to print each text after tokenization.
                Useful for debugging purposes.
            debug (bool): set to True in order to pause after tokenizing
                each text (wait for pressing any key).
                Useful for debugging purposes, if you want to inspect each text
                as is processed.
        """
        self.lowercase = lowercase
        self.debug = debug
        self.verbose = verbose
        colorama.init(autoreset=False, convert=False, strip=False, wrap=True)

        self.pipeline = []

        self.regexes = ExManager().expressions

        if pipeline is None:
            pipeline = self.default_pipeline

        self.build(pipeline)

        self.pipeline.append("(?:\S)")  # CATCH ALL remaining terms
        self.tok = re.compile(r"({})".format("|".join(self.pipeline)))

    def add_to_pipeline(self, term):
        # todo: don't wrap all terms
        self.pipeline.append(self.wrap_non_matching(self.regexes[term]))

    def build(self, pipeline):
        for term in pipeline:
            if term.lower().startswith("emoji"):
                try:
                    re.compile(self.regexes["EMOJI_UCS4"])
                    self.add_to_pipeline("EMOJI_UCS4")
                except re.error:
                    re.compile(self.regexes["EMOJI_UCS2"])
                    self.add_to_pipeline("EMOJI_UCS2")
            else:
                self.add_to_pipeline(term)

    @staticmethod
    def wrap_non_matching(exp):
        return "(?:{})".format(exp)

    def verbose_text(self, text, tokenized):
        # print(text.rstrip())
        for term in tokenized:
            print(colored(term, 'red', attrs=["underline"]), end=" ")
        print()
        if self.debug:
            input()
        else:
            print()

    def tokenize(self, text):
        escaped = html.unescape(text)
        tokenized = self.tok.findall(escaped)

        if self.verbose:
            self.verbose_text(text, tokenized)

        if self.lowercase:
            tokenized = [t.lower() for t in tokenized]

        return tokenized


class SocialTokenizer:
    """
    **Deprecated**

    A parametric tokenizer that understands many expression found in natural
    language such as hashtags, dates, times, emoticons and much more.
    """

    def __init__(self, lowercase=False, verbose=False, debug=False, **kwargs):
        """

        Args:
            lowercase (bool): set to True in order to lowercase the text
            verbose (bool): set to True to print each text after tokenization.
                Useful for debugging purposes.
            debug (bool): set to True in order to pause after tokenizing
                each text (wait for pressing any key).
                Useful for debugging purposes, if you want to inspect each text
                as is processed.

        Kwargs ():
            emojis (bool): True to keep emojis
            urls (bool): True to keep urls
            tags (bool): True to keep tags: <tag>
            emails (bool): True to keep emails
            users (bool): True to keep users handles: @cbaziotis
            hashtags (bool): True to keep hashtags
            cashtags (bool): True to keep cashtags
            phones (bool): True to keep phones
            percents (bool): True to keep percents
            money (bool): True to keep money expressions
            date (bool): True to keep date expressions
            time (bool): True to keep time expressions
            acronyms (bool): True to keep acronyms
            emoticons (bool): True to keep emoticons
            censored (bool): True to keep censored words: f**k
            emphasis (bool): True to keep words with emphasis: *very* good
            numbers (bool): True to keep numbers
        """

        self.lowercase = lowercase
        self.debug = debug
        self.verbose = verbose
        colorama.init(autoreset=False, convert=False, strip=False, wrap=True)
        pipeline = []
        self.regexes = ExManager().expressions

        emojis = kwargs.get("emojis", True)
        urls = kwargs.get("urls", True)
        tags = kwargs.get("tags", True)
        emails = kwargs.get("emails", True)
        users = kwargs.get("users", True)
        hashtags = kwargs.get("hashtags", True)
        cashtags = kwargs.get("cashtags", True)
        phones = kwargs.get("phones", True)
        percents = kwargs.get("percents", True)
        money = kwargs.get("money", True)
        date = kwargs.get("date", True)
        time = kwargs.get("time", True)
        acronyms = kwargs.get("acronyms", True)
        emoticons = kwargs.get("emoticons", True)
        censored = kwargs.get("censored", True)
        emphasis = kwargs.get("emphasis", True)
        numbers = kwargs.get("numbers", True)

        if emojis:
            try:
                re.compile(self.regexes["EMOJI_UCS4"])
                pipeline.append(self.regexes["EMOJI_UCS4"])
            except re.error:
                re.compile(self.regexes["EMOJI_UCS2"])
                pipeline.append(self.regexes["EMOJI_UCS2"])

        if urls:
            pipeline.append(self.regexes["URL"])

        if tags:
            pipeline.append(self.regexes["TAG"])

        if emails:
            pipeline.append(self.wrap_non_matching(self.regexes["EMAIL"]))

        if users:
            pipeline.append(self.wrap_non_matching(self.regexes["USER"]))

        if hashtags:
            pipeline.append(self.wrap_non_matching(self.regexes["HASHTAG"]))

        if cashtags:
            pipeline.append(self.wrap_non_matching(self.regexes["CASHTAG"]))

        if phones:
            pipeline.append(self.wrap_non_matching(self.regexes["PHONE"]))

        if percents:
            pipeline.append(self.wrap_non_matching(self.regexes["PERCENT"]))

        if money:
            pipeline.append(self.wrap_non_matching(self.regexes["MONEY"]))

        if date:
            pipeline.append(self.wrap_non_matching(self.regexes["DATE"]))

        if time:
            pipeline.append(self.wrap_non_matching(self.regexes["TIME"]))

        if acronyms:
            pipeline.append(self.wrap_non_matching(self.regexes["ACRONYM"]))

        if emoticons:
            pipeline.append(self.regexes["LTR_FACE"])
            pipeline.append(self.regexes["RTL_FACE"])

        if censored:
            pipeline.append(self.wrap_non_matching(self.regexes["CENSORED"]))

        if emphasis:
            pipeline.append(self.wrap_non_matching(self.regexes["EMPHASIS"]))

        # terms like 'eco-friendly', 'go_to', 'john's' - maybe remove the ' or add a parameter for it
        # pipeline.append(r"(?:\b[a-zA-Z]+[a-zA-Z'\-_]+[a-zA-Z]+\b)")

        # <3 ^5
        if emoticons:
            pipeline.append(
                self.wrap_non_matching(self.regexes["REST_EMOTICONS"]))

        if numbers:
            pipeline.append(self.regexes["NUMBER"])

        # any other word
        pipeline.append(self.regexes["WORD"])

        # EASTERN EMOTICONS - (^_^;)   (>_<)>  ＼(^o^)／
        if emoticons:
            pipeline.append(
                self.wrap_non_matching(self.regexes["EASTERN_EMOTICONS"]))

        # keep repeated puncts as one term
        # pipeline.append(r"")

        pipeline.append("(?:\S)")  # CATCH ALL remaining terms

        self.tok = re.compile(r"({})".format("|".join(pipeline)))

    @staticmethod
    def wrap_non_matching(exp):
        return "(?:{})".format(exp)

    def verbose_text(self, text, tokenized):
        # print(text.rstrip())
        for term in tokenized:
            print(colored(term, 'red', attrs=["underline"]), end=" ")
        print()
        if self.debug:
            input()
        else:
            print()

    def tokenize(self, text):
        escaped = html.unescape(text)
        tokenized = self.tok.findall(escaped)

        if self.verbose:
            self.verbose_text(text, tokenized)

        if self.lowercase:
            tokenized = [t.lower() for t in tokenized]

        return tokenized

# sentences = []

# [print(s) for s in sentences]
# tokenizer = SocialTokenizer(debug=True, verbose=True)
#
# for s in sentences:
#     tokenizer.tokenize(s)



"""
Created by Christos Baziotis.
"""
from ekphrasis.classes.tokenizer import SocialTokenizer


social_tokenizer = SocialTokenizer(lowercase=False).tokenize

sents = [
    "CANT WAIT for the new season of #TwinPeaks ＼(^o^)／ yaaaay!!! #davidlynch #tvseries :)))",
    "I saw the new #johndoe movie and it suuuuucks!!! WAISTED $10... #badmovies 3:/",
    "@SentimentSymp:  can't wait for the Nov 9 #Sentiment talks!  YAAAAAAY !!! >:-D http://sentimentsymposium.com/.",
]

for s in sents:
    print("SC : ", social_tokenizer(s))  # social tokenizer



import re
from collections import Counter
from difflib import SequenceMatcher
from functools import lru_cache

from ekphrasis.utils.helpers import read_stats

REGEX_TOKEN = re.compile(r'\b[a-z]{2,}\b')


class SpellCorrector:
    """
    The SpellCorrector extends the functionality of the Peter Norvig's
    spell-corrector in http://norvig.com/spell-correct.html
    """

    def __init__(self, corpus="english"):
        """

        :param corpus: the statistics from which corpus to use for the spell correction.
        """
        super().__init__()
        self.WORDS = Counter(read_stats(corpus, 1))
        self.N = sum(self.WORDS.values())

    @staticmethod
    def tokens(text):
        return REGEX_TOKEN.findall(text.lower())

    def P(self, word):
        """
        Probability of `word`.
        """
        return self.WORDS[word] / self.N

    def most_probable(self, words):
        _known = self.known(words)
        if _known:
            return max(_known, key=self.P)
        else:
            return []

    @staticmethod
    def edit_step(word):
        """
        All edits that are one edit away from `word`.
        """
        letters = 'abcdefghijklmnopqrstuvwxyz'
        splits = [(word[:i], word[i:]) for i in range(len(word) + 1)]
        deletes = [L + R[1:] for L, R in splits if R]
        transposes = [L + R[1] + R[0] + R[2:] for L, R in splits if len(R) > 1]
        replaces = [L + c + R[1:] for L, R in splits if R for c in letters]
        inserts = [L + c + R for L, R in splits for c in letters]
        return set(deletes + transposes + replaces + inserts)

    def edits2(self, word):
        """
        All edits that are two edits away from `word`.
        """
        return (e2 for e1 in self.edit_step(word)
                for e2 in self.edit_step(e1))

    def known(self, words):
        """
        The subset of `words` that appear in the dictionary of WORDS.
        """
        return set(w for w in words if w in self.WORDS)

    @staticmethod
    def similar(a, b):
        return SequenceMatcher(None, a, b).ratio()

    def edit_candidates(self, word, assume_wrong=False, fast=True):
        """
        Generate possible spelling corrections for word.
        """

        if fast:
            if assume_wrong:
                return self.known(self.edit_step(word)) or [word]
            else:
                return self.known([word]) or self.known(self.edit_step(word)) or [word]
        else:
            if assume_wrong:
                ttt = self.known(self.edit_step(word)) or self.known(self.edits2(word)) or {word}
                return ttt
            else:
                return self.known([word]) or self.known(self.edit_step(word)) or self.known(self.edits2(word)) or [word]

    # def distance_candidates(self, word, max_distance=3):
    #     """
    #     Generate possible spelling corrections for word.
    #     """
    #     candidates = [w for w in self.WORDS if w]
    #     return self.known([word]) or self.known(self.edit_step(word)) or self.known(self.edits2(word)) or [word]

    @lru_cache(maxsize=None)
    def correct(self, word, assume_wrong=False, fast=False):
        """
        Most probable spelling correction for word.
        """
        return max(self.edit_candidates(word, assume_wrong=assume_wrong, fast=fast), key=self.P)

    def correct_text(self, text):
        """
        Correct all the words within a text, returning the corrected text."""

        return re.sub('[a-zA-Z]+', self.correct_match, text)

    def correct_match(self, match):
        """
        Spell-correct word in match, and preserve proper upper/lower/title case.
        """

        word = match.group()
        return self.case_of(word)(self.correct(word.lower()))

    def correct_word(self, word, assume_wrong=False, fast=False):
        """
        Spell-correct word in match, and preserve proper upper/lower/title case.
        """

        return self.case_of(word)(self.correct(word.lower(), assume_wrong=assume_wrong, fast=fast))

    @staticmethod
    def case_of(text):
        """
        Return the case-function appropriate for text: upper, lower, title, or just str.
        """

        return (str.upper if text.isupper() else
                str.lower if text.islower() else
                str.title if text.istitle() else
                str)

    def elong_normalized_candidates(self, word, acc=None):
        if acc is None:
            acc = []
        candidates = [w for w in set(word) if word.count(w) > 1]
        for c in candidates:
            _w = word.replace(c + c, c)
            if _w in acc:
                continue
            acc.append(_w)
            self.elong_normalized_candidates(_w, acc)
        return acc + [word]

    def best_elong_candidate(self, word):
        candidates = self.elong_normalized_candidates(word)
        best = self.most_probable(candidates)
        return best or word

    def normalize_elongated(self, word):
        return self.case_of(word)(self.best_elong_candidate(word.lower()))



import os

import pickle


def read_slangdict():
    filename = os.path.join(os.path.dirname(__file__), "slangdict.pickle")
    if os.path.isfile(filename):
        print("Reading data...")
        data = pickle.load(open(filename, 'rb'))
        return data



import itertools
import re

from termcolor import cprint

# additional negations: nowhere
negation_words = {"'t", "ain't", 'aint', "aren't", 'arent', 'cant',
                  "didn't", 'didnt', "doesn't", 'doesnt', "don't", 'dont',
                  "hadn't", 'hadnt', "hasn't", 'hasnt', "haven't", 'havent', "isn't", 'isnt',
                  'never', 'no', 'none', 'noone', 'not', 'nothing', 'wont', }
negation_modals = {"couldn't", 'couldnt', "shouldn't", 'shouldnt', "wouldn't", 'wouldnt'}
contrast_words = {"but", "although", "though", "however", "despite", "whereas", "while", "unlike", "still"}
neg_puncts = {"\n", ".", "?", ":", "..."}


def unpack_contractions(text):
    """
    Replace *English* contractions in ``text`` str with their unshortened forms.
    N.B. The "'d" and "'s" forms are ambiguous (had/would, is/has/possessive),
    so are left as-is.

    ---------
    ---------

    Important Note: The function is taken from textacy (https://github.com/chartbeat-labs/textacy).

    See textacy.preprocess.unpack_contractions(text)
    -> http://textacy.readthedocs.io/en/latest/api_reference.html#textacy.preprocess.unpack_contractions


    The reason that textacy is not added as a dependency is to avoid having the user to install it's dependencies (such as SpaCy),
    in order to just use this function.

    """
    # standard
    text = re.sub(r"(\b)([Aa]re|[Cc]ould|[Dd]id|[Dd]oes|[Dd]o|[Hh]ad|[Hh]as|[Hh]ave|[Ii]s|[Mm]ight|[Mm]ust|[Ss]hould|[Ww]ere|[Ww]ould)n't", r"\1\2 not", text)
    text = re.sub(r"(\b)([Hh]e|[Ii]|[Ss]he|[Tt]hey|[Ww]e|[Ww]hat|[Ww]ho|[Yy]ou)'ll", r"\1\2 will", text)
    text = re.sub(r"(\b)([Tt]hey|[Ww]e|[Ww]hat|[Ww]ho|[Yy]ou)'re", r"\1\2 are", text)
    text = re.sub(r"(\b)([Ii]|[Ss]hould|[Tt]hey|[Ww]e|[Ww]hat|[Ww]ho|[Ww]ould|[Yy]ou)'ve", r"\1\2 have", text)
    # non-standard
    text = re.sub(r"(\b)([Cc]a)n't", r"\1\2n not", text)
    text = re.sub(r"(\b)([Ii])'m", r"\1\2 am", text)
    text = re.sub(r"(\b)([Ll]et)'s", r"\1\2 us", text)
    text = re.sub(r"(\b)([Ww])on't", r"\1\2ill not", text)
    text = re.sub(r"(\b)([Ss])han't", r"\1\2hall not", text)
    text = re.sub(r"(\b)([Yy])(?:'all|a'll)", r"\1\2ou all", text)
    return text


def doc_ngrams(doc, n_from=1, n_to=2):
    return list(itertools.chain.from_iterable([[doc[i:i + n] for i in range(len(doc) - (n - 1))]
                                               for n in range(n_from, n_to + 1)]))


def find_negations(doc, neg_comma=True, neg_modals=True, debug=False):
    """
    Takes as input a list of words and returns the positions (indices) of the words
    that are in the context of a negation.

    :param list doc: a list of words (strings)
    :param bool neg_comma: if True, the negation context ends on a comma
    :param bool neg_modals: if True, include negation modals in the set of negation words
    :param bool debug: if True, print the text color coded by context
    :return set: a set of the word positions inside a negation

    """
    doc_context = []
    append = doc_context.append
    negation_stopset = neg_puncts | {","} if neg_comma else set()
    negation_startset = negation_words | negation_modals if neg_modals else set()

    # status == "normal" means outside of parentheses
    # status == "parentheses" means inside parentheses
    # status[XXX] == True means that the context XXX is negated
    # status[XXX] == False means that the context XXX is affirmative
    status = {"normal": False, "parentheses": False}

    # pointer to the current context
    current = "normal"

    for i, tok in enumerate(doc):

        if tok in negation_startset:
            status[current] = True
            if debug:
                cprint(tok, 'red', attrs=['bold'], end=' ')
            continue

        if tok in negation_stopset | contrast_words:
            if debug:
                if status[current]:
                    cprint(tok, 'green', attrs=['bold'], end=' ')
                else:
                    print(tok, end=" ")
            status[current] = False
            continue

        if tok == "(":
            current = "parentheses"
            if debug:
                cprint(tok, 'green', attrs=['bold'], end=' ')
            continue

        if tok == ")":
            status["parentheses"] = False  # in order to be false the next time it goes in to a parentheses
            current = "normal"
            if debug:
                cprint(tok, 'green', attrs=['bold'], end=' ')
            continue

        if debug:
            if status[current]:
                cprint(tok, 'magenta', end=' ')
            else:
                print(tok, end=" ")

        if status[current]:
            append(i)

    if debug:
        print()
    # input("press to continue...")

    return set(doc_context)


def mark_doc(doc, wids, mark=None, pos=None):
    """
    Given a list of words and a set of word positions, mark the words in those positions.
    :param list doc: a list of words (strings)
    :param set wids: the positions of the words to be marked
    :param string mark: a string that sets the mark that will be applied
                        to each of the selected words
    :param string pos: can be one of {"prefix", "suffix"}
    :return: the marked list of words
    """
    if mark is None:
        mark = "NEG"

    if pos is None:
        pos = "suffix"

    marked_doc = []

    for i, tok in enumerate(doc):
        if i in wids:
            if pos == "prefix":
                word = mark + "_" + tok
            else:
                word = tok + "_" + mark
            marked_doc.append(word)
        else:
            marked_doc.append(tok)

    return marked_doc



"""
Created by Christos Baziotis.
"""

demo_sents = [
    "CANT WAIT for the new season of #TwinPeaks ＼(^o^)／ yaaaay!!! #davidlynch #tvseries :)))",
    "I saw the new #johndoe movie and it suuuuucks!!! WAISTED $10... #badmovies >3:/",
    "@SentimentSymp:  can't wait for the Nov 9 #Sentiment talks!  YAAAAAAY !!! >:-D http://sentimentsymposium.com/.",
]



import operator
import os
import pickle
import sys
import zipfile
from functools import reduce
from os import path
from urllib.request import urlretrieve

stats_dir = path.join(os.path.dirname(os.path.dirname(__file__)), 'stats')


def parse_stats(name, sep='\t', ngram_sep='_'):
    """
    Read key,value pairs from file.
    """
    print("reading ngrams", name)
    d = {}
    with open(name, "r", encoding="utf-8") as f:
        for line in f:
            values = line.split(sep)
            if len(values) > 2:
                d[ngram_sep.join(values[:-1])] = int(values[-1])
            else:
                d[values[0]] = int(values[1])

    return d


def read_stats(corpus, ngram):
    check_stats_files()
    print("Reading " + "{} - {}grams ...".format(corpus, ngram))
    text = path.join(*[stats_dir, corpus, "counts_{}grams.txt".format(ngram)])
    pickled = path.join(
        *[stats_dir, corpus, "counts_{}grams.pickle".format(ngram)])

    if os.path.isfile(pickled):
        with open(pickled, "rb") as f:
            stats = pickle.load(f)
            return stats
    elif os.path.isfile(text):
        print("generating binary file for faster loading...")
        stats = parse_stats(text)
        with open(pickled, "wb") as f:
            pickle.dump(stats, f)
        return stats
    else:
        print("stats file not available!")
        sys.exit(1)


def listdir_nohidden(path):
    return [f for f in os.listdir(path) if not f.startswith('.')]


def check_stats_files():
    if not os.path.exists(stats_dir) or len(listdir_nohidden(stats_dir)) == 0:
        print("Word statistics files not found!\nDownloading...", end=" ")
        url = "https://www.dropbox.com/s/a84otqrg6u1c5je/stats.zip?dl=1"
        urlretrieve(url, "stats.zip")
        print("done!")

        print("Unpacking...", end=" ")
        with zipfile.ZipFile("stats.zip", "r") as zip_ref:
            zip_ref.extractall(stats_dir)

        os.remove("stats.zip")
        print("done!")


def product(nums):
    """
    Return the product of a sequence of numbers.
    """
    return reduce(operator.mul, nums, 1)


check_stats_files()
